"""Sphinx extension that adds Run and Edit buttons to code blocks using Pyodide."""

from __future__ import annotations

from sphinx_pyodide_runner.extension import __version__, setup

__all__ = ["__version__", "setup"]
